"""
Quantum circuit: the state simulator.
Supports qubit (dim=2) and qudit (3 <= dim <= 36) systems.
For string-encoded samples/counts, digits use 0-9A-Z where A=10, ..., Z=35.
"""

# pylint: disable=invalid-name

from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple
from functools import reduce
from operator import add

import numpy as np
import tensornetwork as tn

from . import gates
from . import channels
from .cons import backend, contractor, dtypestr, npdtype, _ALPHABET
from .quantum import QuOperator, identity, _infer_num_sites, onehot_d_tensor
from .simplify import _full_light_cone_cancel
from .basecircuit import BaseCircuit

Gate = gates.Gate
Tensor = Any


class Circuit(BaseCircuit):
    r"""
    ``Circuit`` class.
    Simple usage demo below.

    .. code-block:: python

        c = tc.Circuit(3)
        c.H(1)
        c.CNOT(0, 1)
        c.RX(2, theta=tc.num_to_tensor(1.))
        c.expectation([tc.gates.z(), (2, )]) # 0.54

    """

    is_dm = False

    def __init__(
        self,
        nqubits: int,
        inputs: Optional[Tensor] = None,
        mps_inputs: Optional[QuOperator] = None,
        split: Optional[Dict[str, Any]] = None,
        dim: Optional[int] = None,
    ) -> None:
        r"""
        Circuit object based on state simulator.
        Do not use this class with d!=2 directly, use tc.QuditCircuit instead for qudit systems.

        :param nqubits: The number of qubits in the circuit.
        :type nqubits: int
        :param dim: The local Hilbert space dimension per site. Qudit is supported for 2 <= d <= 36.
        :type dim: If None, the dimension of the circuit will be `2`, which is a qubit system.
        :param inputs: If not None, the initial state of the circuit is taken as ``inputs``
            instead of :math:`\vert 0 \rangle^n` qubits, defaults to None.
        :type inputs: Optional[Tensor], optional
        :param mps_inputs: QuVector for a MPS like initial wavefunction.
        :type mps_inputs: Optional[QuOperator]
        :param split: dict if two qubit gate is ready for split, including parameters for at least one of
            ``max_singular_values`` and ``max_truncation_err``.
        :type split: Optional[Dict[str, Any]]
        """
        self._d = 2 if dim is None else dim
        self.inputs = inputs
        self.mps_inputs = mps_inputs
        self.split = split
        self._nqubits = nqubits

        self.circuit_param = {
            "nqubits": nqubits,
            "inputs": inputs,
            "mps_inputs": mps_inputs,
            "split": split,
            "dim": dim,
        }
        if (inputs is None) and (mps_inputs is None):
            nodes = self.all_zero_nodes(nqubits, dim=self._d)
            self._front = [n.get_edge(0) for n in nodes]
        elif inputs is not None:  # provide input function
            inputs = backend.convert_to_tensor(inputs)
            inputs = backend.cast(inputs, dtype=dtypestr)
            inputs = backend.reshape(inputs, [-1])
            N = inputs.shape[0]
            n = _infer_num_sites(N, dim=self._d)
            assert n == nqubits or n == 2 * nqubits
            inputs = backend.reshape(inputs, [self._d for _ in range(n)])
            inputs = Gate(inputs)
            nodes = [inputs]
            self._front = [inputs.get_edge(i) for i in range(n)]
        else:  # mps_inputs is not None
            mps_nodes = list(mps_inputs.nodes)  # type: ignore
            for i, n in enumerate(mps_nodes):
                mps_nodes[i].tensor = backend.cast(n.tensor, dtypestr)  # type: ignore
            mps_edges = mps_inputs.out_edges + mps_inputs.in_edges  # type: ignore
            ndict, edict = tn.copy(mps_nodes)
            new_nodes = []
            for n in mps_nodes:
                new_nodes.append(ndict[n])
            new_front = []
            for e in mps_edges:
                new_front.append(edict[e])
            nodes = new_nodes
            self._front = new_front

        self.coloring_nodes(nodes)
        self._nodes = nodes

        self._start_index = len(nodes)
        # self._start = nodes
        # self._meta_apply()

        # self._qcode = ""  # deprecated
        # self._qcode += str(self._nqubits) + "\n"
        self._qir: List[Dict[str, Any]] = []
        self._extra_qir: List[Dict[str, Any]] = []

    def replace_mps_inputs(self, mps_inputs: QuOperator) -> None:
        """
        Replace the input state in MPS representation while keep the circuit structure unchanged.

        :Example:
        >>> c = tc.Circuit(2)
        >>> c.X(0)
        >>>
        >>> c2 = tc.Circuit(2, mps_inputs=c.quvector())
        >>> c2.X(0)
        >>> c2.wavefunction()
        array([1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j], dtype=complex64)
        >>>
        >>> c3 = tc.Circuit(2)
        >>> c3.X(0)
        >>> c3.replace_mps_inputs(c.quvector())
        >>> c3.wavefunction()
        array([1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j], dtype=complex64)

        :param mps_inputs: (Nodes, dangling Edges) for a MPS like initial wavefunction.
        :type mps_inputs: Tuple[Sequence[Gate], Sequence[Edge]]
        """
        mps_nodes = mps_inputs.nodes
        mps_edges = mps_inputs.out_edges + mps_inputs.in_edges
        ndict, edict = tn.copy(mps_nodes)
        new_nodes = []
        for n in mps_nodes:
            new_nodes.append(ndict[n])
        new_front = []
        for e in mps_edges:
            new_front.append(edict[e])
        old = set(id(n) for n in self._nodes[: self._start_index])
        j = -1
        for n in self._nodes[: self._start_index]:
            for e in n:
                if e.is_dangling():
                    j += 1
                    self._front[j] = new_front[j]
                else:
                    if (id(e.node1) in old) and (id(e.node2) in old):
                        pass
                    else:
                        j += 1
                        if id(e.node2) == id(n):
                            other = (e.node1, e.axis1)
                        else:  # id(e.node1) == id(n):
                            other = (e.node2, e.axis2)
                        e.disconnect()
                        new_front[j] ^ other[0][other[1]]
        j += 1
        self._front += new_front[j:]
        self.coloring_nodes(new_nodes)
        self._nodes = new_nodes + self._nodes[self._start_index :]
        self._start_index = len(new_nodes)

    # TODO(@refraction-ray): add noise support in IR
    # TODO(@refraction-ray): unify mid measure to basecircuit

    def mid_measurement(self, index: int, keep: int = 0) -> Tensor:
        """
        Middle measurement in z-basis on the circuit, note the wavefunction output is not normalized
        with ``mid_measurement`` involved, one should normalize the state manually if needed.
        This is a post-selection method as keep is provided as a prior.

        :param index: The index of qubit that the Z direction postselection applied on.
        :type index: int
        :param keep: the post-selected digit in {0, ..., d-1}, defaults to be 0.
        :type keep: int, optional
        """
        # normalization not guaranteed
        gate = np.array(
            [[0.0] if _idx != keep else [1.0] for _idx in range(self._d)],
            dtype=npdtype,
        )

        mg1 = tn.Node(gate)
        mg2 = tn.Node(gate)
        # mg1.flag = "post-select"
        # mg1.is_dagger = False
        # mg1.id = id(mg1)
        # mg2.flag = "post-select"
        # mg2.is_dagger = False
        # mg2.id = id(mg2)
        self.coloring_nodes([mg1, mg2], flag="post-select")
        mg1.get_edge(0) ^ self._front[index]
        mg1.get_edge(1) ^ mg2.get_edge(1)
        self._front[index] = mg2.get_edge(0)
        self._nodes.append(mg1)
        self._nodes.append(mg2)
        r = backend.convert_to_tensor(keep)
        r = backend.cast(r, "int32")
        return r

    mid_measure = mid_measurement
    post_select = mid_measurement
    post_selection = mid_measurement

    def depolarizing2(
        self,
        index: int,
        *,
        px: float,
        py: float,
        pz: float,
        status: Optional[float] = None,
    ) -> float:
        """
        Apply a depolarizing channel to the circuit in a Monte Carlo way.
        For each call, one of the Pauli gates (X, Y, Z) or an Identity gate is applied to the qubit
        at the given index based on the probabilities `px`, `py`, and `pz`.

        :param index: The index of the qubit to apply the depolarizing channel on.
        :type index: int
        :param px: The probability of applying an X gate.
        :type px: float
        :param py: The probability of applying a Y gate.
        :type py: float
        :param pz: The probability of applying a Z gate.
        :type pz: float
        :param status: A random number between 0 and 1 to determine which gate to apply. If None,
                       a random number is generated automatically. Defaults to None.
        :type status: Optional[float], optional
        :return: Returns 0.0. The function modifies the circuit in place.
        :rtype: float
        """
        if status is None:
            status = backend.implicit_randu()[0]
        g = backend.cond(
            status < px,
            lambda: gates.x().tensor,  # type: ignore
            lambda: backend.cond(
                status < px + py,  # type: ignore
                lambda: gates.y().tensor,  # type: ignore
                lambda: backend.cond(
                    status < px + py + pz,  # type: ignore
                    lambda: gates.z().tensor,  # type: ignore
                    lambda: gates.i().tensor,  # type: ignore
                ),
            ),
        )
        # after implementing this, I realized that plain if is enough here for jit
        # the failure for previous implementation is because we use self.X(i) inside ``if``,
        # which has list append and incur bug in tensorflow jit
        # in terms of jax jit, the only choice is jax.lax.cond, since ``if tensor``` paradigm
        # is not supported in jax jit at all. (``Concrete Tensor Error``)
        self.any(index, unitary=g)  # type: ignore
        return 0.0
        # roughly benchmark shows that performance of two depolarizing in terms of
        # building time and running time are similar

    # overwritten now, deprecated
    def depolarizing_reference(
        self,
        index: int,
        *,
        px: float,
        py: float,
        pz: float,
        status: Optional[float] = None,
    ) -> Tensor:
        """
        Apply depolarizing channel in a Monte Carlo way,
        i.e. for each call of this method, one of gates from
        X, Y, Z, I are applied on the circuit based on the probability
        indicated by ``px``, ``py``, ``pz``.

        :param index: The qubit that depolarizing channel is on
        :type index: int
        :param px: probability for X noise
        :type px: float
        :param py: probability for Y noise
        :type py: float
        :param pz: probability for Z noise
        :type pz: float
        :param status: random seed uniformly from 0 to 1, defaults to None (generated implicitly)
        :type status: Optional[float], optional
        :return: int Tensor, the element lookup: [0: x, 1: y, 2: z, 3: I]
        :rtype: Tensor
        """

        # px/y/z here not support differentiation for now
        # jit compatible for now
        # assert px + py + pz < 1 and px >= 0 and py >= 0 and pz >= 0

        def step_function(x: Tensor) -> Tensor:
            r = (
                backend.sign(x - px)
                + backend.sign(x - px - py)
                + backend.sign(x - px - py - pz)
            )
            r = backend.cast(r / 2 + 1.5, dtype="int32")
            # [0: x, 1: y, 2: z, 3: I]

            return r

        if status is None:
            status = backend.implicit_randu()[0]
        r = step_function(status)
        rv = backend.onehot(r, 4)
        rv = backend.cast(rv, dtype=dtypestr)
        g = (
            rv[0] * gates.x().tensor  # type: ignore
            + rv[1] * gates.y().tensor  # type: ignore
            + rv[2] * gates.z().tensor  # type: ignore
            + rv[3] * gates.i().tensor  # type: ignore
        )
        self.any(index, unitary=g)  # type: ignore
        return r

    def unitary_kraus2(
        self,
        kraus: Sequence[Gate],
        *index: int,
        prob: Optional[Sequence[float]] = None,
        status: Optional[float] = None,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Apply a unitary Kraus channel to the circuit using a Monte Carlo approach. This method is functionally
        similar to `unitary_kraus` but uses `backend.switch` for selecting the Kraus operator, which can have
        different performance characteristics on some backends.

        A random Kraus operator from the provided list is applied to the circuit based on the given probabilities.
        This method is jittable and suitable for simulating noisy quantum circuits where the noise is represented
        by unitary Kraus operators.

        .. warning::
            This method may have issues with `vmap` due to potential concurrent access locks, potentially related with
            `backend.switch`. `unitary_kraus` is generally recommended.

        :param kraus: A sequence of `Gate` objects representing the unitary Kraus operators.
        :type kraus: Sequence[Gate]
        :param index: The qubit indices on which to apply the Kraus channel.
        :type index: int
        :param prob: A sequence of probabilities corresponding to each Kraus operator. If None, probabilities
                     are derived from the operators themselves. Defaults to None.
        :type prob: Optional[Sequence[float]], optional
        :param status: A random number between 0 and 1 to determine which Kraus operator to apply. If None,
                       a random number is generated automatically. Defaults to None.
        :type status: Optional[float], optional
        :param name: An optional name for the operation. Defaults to None.
        :type name: Optional[str], optional
        :return: A tensor indicating which Kraus operator was applied.
        :rtype: Tensor
        """

        # dont use, has issue conflicting with vmap, concurrent access lock emerged
        # potential issue raised from switch
        # general impl from Monte Carlo trajectory depolarizing above
        # still jittable
        # speed is similar to ``unitary_kraus``
        def index2gate2(r: Tensor, kraus: Sequence[Tensor]) -> Tensor:
            # r is int type Tensor of shape []
            return backend.switch(r, [lambda _=k: _ for k in kraus])  # type: ignore

        return self._unitary_kraus_template(
            kraus,
            *index,
            prob=prob,
            status=status,
            get_gate_from_index=index2gate2,
            name=name,
        )

    def unitary_kraus(
        self,
        kraus: Sequence[Gate],
        *index: int,
        prob: Optional[Sequence[float]] = None,
        status: Optional[float] = None,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Apply unitary gates in ``kraus`` randomly based on corresponding ``prob``.
        If ``prob`` is ``None``, this is reduced to kraus channel language.

        :param kraus: List of ``tc.gates.Gate`` or just Tensors
        :type kraus: Sequence[Gate]
        :param prob: prob list with the same size as ``kraus``, defaults to None
        :type prob: Optional[Sequence[float]], optional
        :param status: random seed between 0 to 1, defaults to None
        :type status: Optional[float], optional
        :return: shape [] int dtype tensor indicates which kraus gate is actually applied
        :rtype: Tensor
        """
        # general impl from Monte Carlo trajectory depolarizing above
        # still jittable

        def index2gate(r: Tensor, kraus: Sequence[Tensor]) -> Tensor:
            # r is int type Tensor of shape []
            l = len(kraus)
            r = backend.onehot(r, l)
            r = backend.cast(r, dtype=dtypestr)
            return reduce(add, [r[i] * kraus[i] for i in range(l)])

        return self._unitary_kraus_template(
            kraus,
            *index,
            prob=prob,
            status=status,
            get_gate_from_index=index2gate,
            name=name,
        )

    def _unitary_kraus_template(
        self,
        kraus: Sequence[Gate],
        *index: int,
        prob: Optional[Sequence[float]] = None,
        status: Optional[float] = None,
        get_gate_from_index: Optional[
            Callable[[Tensor, Sequence[Tensor]], Tensor]
        ] = None,
        name: Optional[str] = None,
    ) -> Tensor:  # DRY
        sites = len(index)
        kraus = [k.tensor if isinstance(k, tn.Node) else k for k in kraus]
        kraus = [gates.array_to_tensor(k) for k in kraus]
        kraus = [backend.reshapem(k) for k in kraus]
        if prob is None:
            prob = [
                backend.real(backend.trace(backend.adjoint(k) @ k) / k.shape[0])
                for k in kraus
            ]
            kraus = [
                k / backend.cast(backend.sqrt(p), dtypestr) for k, p in zip(kraus, prob)
            ]
        if not backend.is_tensor(prob):
            prob = backend.convert_to_tensor(prob)
        prob_cumsum = backend.cumsum(prob)
        l = int(prob.shape[0])  # type: ignore

        def step_function(x: Tensor) -> Tensor:
            if l == 1:
                r = backend.convert_to_tensor(0.0)
            else:
                r = backend.sum(
                    backend.stack(
                        [backend.sign(x - prob_cumsum[i]) for i in range(l - 1)]
                    )
                )
            r = backend.cast(r / 2.0 + (l - 1) / 2.0, dtype="int32")
            # [0: kraus[0], 1: kraus[1]...]
            return r

        if status is None:
            status = backend.implicit_randu()[0]
        status = backend.convert_to_tensor(status)
        status = backend.real(status)
        prob_cumsum = backend.cast(prob_cumsum, dtype=status.dtype)  # type: ignore
        r = step_function(status)
        if get_gate_from_index is None:
            raise ValueError("no `get_gate_from_index` implementation is provided")
        g = get_gate_from_index(r, kraus)
        g = backend.reshape(g, [self._d for _ in range(sites * 2)])
        self.any(*index, unitary=g, name=name, dim=self._d)  # type: ignore
        return r

    def _general_kraus_tf(
        self,
        kraus: Sequence[Gate],
        *index: int,
        status: Optional[float] = None,
    ) -> float:
        # the graph building time is frustratingly slow, several minutes
        # though running time is in terms of ms
        sites = len(index)
        kraus_tensor = [k.tensor for k in kraus]
        kraus_tensor_f = [lambda _=k: _ for k in kraus_tensor]
        # must return tensor instead of ``tn.Node`` for switch`

        def calculate_kraus_p(i: Tensor) -> Tensor:
            # i: Tensor as int of shape []
            newnodes, newfront = self._copy()  # TODO(@refraction-ray): support reuse?
            # simply reuse=True is wrong, as the circuit is contracting at building
            # self._copy seems slower than self._copy_state, but anyway the building time is unacceptable
            lnewnodes, lnewfront = self._copy(conj=True)
            kraus_i = backend.switch(i, kraus_tensor_f)
            k = gates.Gate(kraus_i)
            kc = gates.Gate(backend.conj(kraus_i))
            # begin connect
            for ind, j in enumerate(index):
                newfront[j] ^ k[ind + sites]
                k[ind] ^ kc[ind]
                kc[ind + sites] ^ lnewfront[j]
            for j in range(self._nqubits):
                if j not in index:
                    newfront[j] ^ lnewfront[j]
            norm_square = contractor(newnodes + lnewnodes + [k, kc]).tensor
            return backend.real(norm_square)

        if status is None:
            status = backend.implicit_randu()[0]

        import tensorflow as tf  # tf only implementation

        weight = 1.0
        fallback_weight = 0.0
        fallback_weight_i = 0
        len_kraus = len(kraus)
        for i in tf.range(len_kraus):  # breaks backend agnostic
            # nested for and if, if tensor inner must come with for in tensor outer, s.t. autograph works
            weight = calculate_kraus_p(i)
            if weight > fallback_weight:
                fallback_weight_i = i
                fallback_weight = weight
            status -= weight
            if status < 0:
                # concern here, correctness not sure in tf jit, fail anyway in jax jit
                break
        # placing a Tensor-dependent break, continue or return inside a Python loop
        # https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/autograph/g3doc/reference/common_errors.md

        if (
            status >= 0 or weight == 0
        ):  # the same concern, but this simple if is easy to convert to ``backend.cond``
            # Floating point error resulted in a malformed sample.
            # Fall back to the most likely case.
            # inspired from cirq implementation (Apcache 2).
            weight = fallback_weight
            i = fallback_weight_i
        kraus_i = backend.switch(i, kraus_tensor_f)
        newgate = kraus_i / backend.cast(backend.sqrt(weight), dtypestr)
        self.any(*index, unitary=newgate)  # type: ignore
        return 0.0

    def _general_kraus_2(
        self,
        kraus: Sequence[Gate],
        *index: int,
        status: Optional[float] = None,
        with_prob: bool = False,
        name: Optional[str] = None,
    ) -> Tensor:
        # the graph building time is frustratingly slow, several minutes
        # though running time is in terms of ms
        # raw running time in terms of s
        # note jax gpu building time is fast, in the order of 10s.!!
        # the typical scenario we are talking: 10 qubits, 3 layers of entangle gates and 3 layers of noise
        # building for jax+GPU ~100s 12 qubit * 5 layers
        # 370s 14 qubit * 7 layers, 0.35s running on vT4
        # vmap, grad, vvag are all fine for this function
        # layerwise jit technique can greatly boost the staging time, see in /examples/mcnoise_boost.py
        sites = len(index)
        kraus_tensor = [k.tensor if isinstance(k, tn.Node) else k for k in kraus]
        kraus_tensor = [gates.array_to_tensor(k) for k in kraus_tensor]

        # tn with hole
        newnodes, newfront = self._copy()
        lnewnodes, lnewfront = self._copy(conj=True)
        des = [newfront[j] for j in index] + [lnewfront[j] for j in index]
        for j in range(self._nqubits):
            if j not in index:
                newfront[j] ^ lnewfront[j]
        ns = contractor(newnodes + lnewnodes, output_edge_order=des)
        ntensor = ns.tensor
        # ns, des

        def calculate_kraus_p(i: int) -> Tensor:
            # i: Tensor as int of shape []
            # kraus_i = backend.switch(i, kraus_tensor_f)
            kraus_i = kraus_tensor[i]
            dm = gates.Gate(ntensor)
            k = gates.Gate(kraus_i)
            kc = gates.Gate(backend.conj(kraus_i))
            # begin connect
            for ind in range(sites):
                dm[ind] ^ k[ind + sites]
                k[ind] ^ kc[ind]
                kc[ind + sites] ^ dm[ind + sites]
            norm_square = contractor([dm, k, kc]).tensor
            return backend.real(norm_square)

        prob = [calculate_kraus_p(i) for i in range(len(kraus))]
        eps = 1e-10
        new_kraus = [
            k / backend.cast(backend.sqrt(w) + eps, dtypestr)
            for w, k in zip(prob, kraus_tensor)
        ]
        pick = self.unitary_kraus(
            new_kraus,
            *index,
            prob=prob,
            status=status,
            name=name,
        )
        if not with_prob:
            return pick
        else:
            return pick, prob

    def general_kraus(
        self,
        kraus: Sequence[Gate],
        *index: int,
        status: Optional[float] = None,
        with_prob: bool = False,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Monte Carlo trajectory simulation of general Kraus channel whose Kraus operators cannot be
        amplified to unitary operators. For unitary operators composed Kraus channel, :py:meth:`unitary_kraus`
        is much faster.

        This function is jittable in theory. But only jax+GPU combination is recommended for jit
        since the graph building time is too long for other backend options; though the running
        time of the function is very fast for every case.

        :param kraus: A list of ``tn.Node`` for Kraus operators.
        :type kraus: Sequence[Gate]
        :param index: The qubits index that Kraus channel is applied on.
        :type index: int
        :param status: Random tensor uniformly between 0 or 1, defaults to be None,
            when the random number will be generated automatically
        :type status: Optional[float], optional
        """
        return self._general_kraus_2(
            kraus,
            *index,
            status=status,
            with_prob=with_prob,
            name=name,
        )

    apply_general_kraus = general_kraus

    @staticmethod
    def apply_general_kraus_delayed(
        krausf: Callable[..., Sequence[Gate]], is_unitary: bool = False
    ) -> Callable[..., None]:
        def apply(
            self: "Circuit",
            *index: int,
            status: Optional[float] = None,
            name: Optional[str] = None,
            **vars: float,
        ) -> None:
            kraus = krausf(**vars)
            if not is_unitary:
                self.apply_general_kraus(kraus, *index, status=status, name=name)
            else:
                self.unitary_kraus(kraus, *index, status=status, name=name)

        return apply

    @classmethod
    def _meta_apply_channels(cls) -> None:
        for k in channels.channels:
            if k in ["depolarizing", "generaldepolarizing"]:
                is_unitary = True
            else:
                is_unitary = False
            setattr(
                cls,
                k,
                cls.apply_general_kraus_delayed(
                    getattr(channels, k + "channel"), is_unitary=is_unitary
                ),
            )
            doc = """
            Apply %s quantum channel on the circuit.
            See :py:meth:`tensorcircuit.channels.%schannel`

            :param index: Site index that the gate applies on.
            :type index: int.
            :param status: uniform external random number between 0 and 1
            :type status: Tensor
            :param vars: Parameters for the channel.
            :type vars: float.
            """ % (
                k,
                k,
            )
            getattr(cls, k).__doc__ = doc

    def is_valid(self) -> bool:
        """
        [WIP], check whether the circuit is legal.

        :return: The bool indicating whether the circuit is legal
        :rtype: bool
        """
        try:
            assert len(self._front) == self._nqubits
            for n in self._nodes:
                for e in n.get_all_dangling():
                    assert e in self._front
            return True
        except AssertionError:
            return False

    def wavefunction(self, form: str = "default") -> tn.Node.tensor:
        """
        Compute the output wavefunction from the circuit.

        :param form: The str indicating the form of the output wavefunction.
            "default": [-1], "ket": [-1, 1], "bra": [1, -1]
        :type form: str, optional
        :return: Tensor with the corresponding shape.
        :rtype: Tensor
        """
        nodes, d_edges = self._copy()
        t = contractor(nodes, output_edge_order=d_edges)
        if form == "default":
            shape = [-1]
        elif form == "ket":
            shape = [-1, 1]
        elif form == "bra":  # no conj here
            shape = [1, -1]
        return backend.reshape(t.tensor, shape=shape)

    state = wavefunction

    def get_quoperator(self) -> QuOperator:
        """
        Get the ``QuOperator`` MPO like representation of the circuit unitary without contraction.

        :return: ``QuOperator`` object for the circuit unitary (open indices for the input state)
        :rtype: QuOperator
        """
        mps = identity([self._d for _ in range(self._nqubits)])
        c = Circuit(self._nqubits, dim=self._d)
        ns, es = self._copy()
        c._nodes = ns
        c._front = es
        c.replace_mps_inputs(mps)
        return QuOperator(c._front[: self._nqubits], c._front[self._nqubits :])

    quoperator = get_quoperator
    # both are not good names, but for backward compatibility

    get_circuit_as_quoperator = get_quoperator
    get_state_as_quvector = BaseCircuit.quvector

    def matrix(self) -> Tensor:
        """
        Get the unitary matrix for the circuit irrespective with the circuit input state.

        :return: The circuit unitary matrix
        :rtype: Tensor
        """
        mps = identity([self._d for _ in range(self._nqubits)])
        c = Circuit(self._nqubits, dim=self._d)
        ns, es = self._copy()
        c._nodes = ns
        c._front = es
        c.replace_mps_inputs(mps)
        return backend.reshapem(c.state())

    def measure_reference(
        self, *index: int, with_prob: bool = False
    ) -> Tuple[str, float]:
        """
        Take measurement on the given quantum lines by ``index``.

        Return format:
        - For d <= 36, the sample is a base-d string using 0-9A-Z (A=10,...).

        :Example:

        >>> c = tc.Circuit(3)
        >>> c.H(0)
        >>> c.h(1)
        >>> c.toffoli(0, 1, 2)
        >>> c.measure(2)
        ('1', -1.0)
        >>> # Another possible output: ('0', -1.0)
        >>> c.measure(2, with_prob=True)
        ('1', (0.25000011920928955+0j))
        >>> # Another possible output: ('0', (0.7499998807907104+0j))

        :param index: Measure on which quantum line.
        :param with_prob: If true, theoretical probability is also returned.
        :return: The sample output and probability (optional) of the quantum line.
        :rtype: Tuple[str, float]
        """
        # not jit compatible due to random number generations!
        sample = ""
        p = 1.0
        for j in index:
            nodes1, edge1 = self._copy()
            nodes2, edge2 = self._copy(conj=True)
            for i, e in enumerate(edge1):
                if i != j:
                    e ^ edge2[i]
            for i in range(len(sample)):
                m = onehot_d_tensor(sample[i], d=self._d)
                nodes1.append(tn.Node(m))
                nodes1[-1].get_edge(0) ^ edge1[index[i]]
                nodes2.append(tn.Node(m))
                nodes2[-1].get_edge(0) ^ edge2[index[i]]
            nodes1.extend(nodes2)
            rho = (
                1
                / p
                * contractor(nodes1, output_edge_order=[edge1[j], edge2[j]]).tensor
            )
            probs = backend.real(backend.diagonal(rho))
            probs /= backend.sum(probs)
            outcome = backend.implicit_randc(self._d, shape=1, p=probs)

            sample += _ALPHABET[outcome]
            p *= float(probs[outcome])

        if with_prob:
            return sample, p
        else:
            return sample, -1.0

    # TODO(@refraction-ray): more _before function like state_before? and better API?

    def expectation(
        self,
        *ops: Tuple[tn.Node, List[int]],
        reuse: bool = True,
        enable_lightcone: bool = False,
        noise_conf: Optional[Any] = None,
        nmc: int = 1000,
        status: Optional[Tensor] = None,
        **kws: Any,
    ) -> Tensor:
        """
        Compute the expectation of corresponding operators.
        For qudit (d > 2),
        ensure that operator tensor shapes are consistent with d (each site contributes two axes of size d).

        Noise shorthand (via noise_conf) is qubit-only; for d>2, use explicit operators.

        :Example:

        >>> c = tc.Circuit(2)
        >>> c.H(0)
        >>> c.expectation((tc.gates.z(), [0]))
        array(0.+0.j, dtype=complex64)

        >>> c = tc.Circuit(2)
        >>> c.cnot(0, 1)
        >>> c.rx(0, theta=0.4)
        >>> c.rx(1, theta=0.8)
        >>> c.h(0)
        >>> c.h(1)
        >>> error1 = tc.channels.generaldepolarizingchannel(0.1, 1)
        >>> error2 = tc.channels.generaldepolarizingchannel(0.06, 2)
        >>> noise_conf = NoiseConf()
        >>> noise_conf.add_noise("rx", error1)
        >>> noise_conf.add_noise("cnot", [error2], [[0, 1]])
        >>> c.expectation((tc.gates.x(), [0]), noise_conf=noise_conf, nmc=10000)
        (0.46274087-3.764033e-09j)

        :param ops: Operator and its position on the circuit,
            eg. ``(tc.gates.z(), [1, ]), (tc.gates.x(), [2, ])`` is for operator :math:`Z_1X_2`.
        :type ops: Tuple[tn.Node, List[int]]
        :param reuse: If True, then the wavefunction tensor is cached for further expectation evaluation,
            defaults to be true.
        :type reuse: bool, optional
        :param enable_lightcone: whether enable light cone simplification, defaults to False
        :type enable_lightcone: bool, optional
        :param noise_conf: Noise Configuration, defaults to None
        :type noise_conf: Optional[NoiseConf], optional
        :param nmc: repetition time for Monte Carlo sampling for noisfy calculation, defaults to 1000
        :type nmc: int, optional
        :param status: external randomness given by tensor uniformly from [0, 1], defaults to None,
            used for noisy circuit sampling
        :type status: Optional[Tensor], optional
        :raises ValueError: "Cannot measure two operators in one index"
        :return: Tensor with one element
        :rtype: Tensor
        """
        if noise_conf is None:
            # if not reuse:
            #     nodes1, edge1 = self._copy()
            #     nodes2, edge2 = self._copy(conj=True)
            # else:  # reuse

            # self._nodes = nodes1
            if enable_lightcone:
                reuse = False
            nodes1 = self.expectation_before(*ops, reuse=reuse)
            if enable_lightcone:
                nodes1 = _full_light_cone_cancel(nodes1)
            return contractor(nodes1).tensor
        else:
            from .noisemodel import expectation_noisfy

            return expectation_noisfy(
                self,
                *ops,
                noise_conf=noise_conf,
                nmc=nmc,
                status=status,
                **kws,
            )


Circuit._meta_apply()
Circuit._meta_apply_channels()


def expectation(
    *ops: Tuple[tn.Node, List[int]],
    ket: Tensor,
    bra: Optional[Tensor] = None,
    conj: bool = True,
    normalization: bool = False,
    dim: Optional[int] = None,
) -> Tensor:
    """
    Compute :math:`\\langle bra\\vert ops \\vert ket\\rangle`.
    For qudit systems (d>2), ops must be reshaped with per-site axes of length d.

    Example 1 (:math:`bra` is same as :math:`ket`)

    >>> c = tc.Circuit(3)
    >>> c.H(0)
    >>> c.ry(1, theta=tc.num_to_tensor(0.8 + 0.7j))
    >>> c.cnot(1, 2)
    >>> state = c.wavefunction() # the state of this circuit
    >>> x1z2 = [(tc.gates.x(), [0]), (tc.gates.z(), [1])] # input qubits
    >>>
    >>> # Expection of this circuit / <state|*x1z2|state>
    >>> c.expectation(*x1z2)
    array(0.69670665+0.j, dtype=complex64)
    >>> tc.expectation(*x1z2, ket=state)
    (0.6967066526412964+0j)
    >>>
    >>> # Normalize(expection of Circuit) / Normalize(<state|*x1z2|state>)
    >>> c.expectation(*x1z2) / tc.backend.norm(state) ** 2
    (0.5550700389340034+0j)
    >>> tc.expectation(*x1z2, ket=state, normalization=True)
    (0.55507004+0j)

    Example 2 (:math:`bra` is different from :math:`ket`)

    >>> c = tc.Circuit(2)
    >>> c.X(1)
    >>> s1 = c.state()
    >>> c2 = tc.Circuit(2)
    >>> c2.X(0)
    >>> s2 = c2.state()
    >>> c3 = tc.Circuit(2)
    >>> c3.H(1)
    >>> s3 = c3.state()
    >>> x1x2 = [(tc.gates.x(), [0]), (tc.gates.x(), [1])]
    >>>
    >>> tc.expectation(*x1x2, ket=s1, bra=s2)
    (1+0j)
    >>> tc.expectation(*x1x2, ket=s3, bra=s2)
    (0.7071067690849304+0j) # 1/sqrt(2)

    :param ket: :math:`ket`. The state in tensor or ``QuVector`` format
    :type ket: Tensor
    :param bra: :math:`bra`, defaults to None, which is the same as ``ket``.
    :type bra: Optional[Tensor], optional
    :param dim: dimension of the circuit (defaults to 2)
    :type dim: int, optional
    :param conj: :math:`bra` changes to the adjoint matrix of :math:`bra`, defaults to True.
    :type conj: bool, optional
    :param normalization: Normalize the :math:`ket` and :math:`bra`, defaults to False.
    :type normalization: bool, optional
    :raises ValueError: "Cannot measure two operators in one index"
    :return: The result of :math:`\\langle bra\\vert ops \\vert ket\\rangle`.
    :rtype: Tensor
    """
    dim = 2 if dim is None else dim
    if bra is None:
        bra = ket
    if isinstance(ket, QuOperator):
        if conj is True:
            bra = bra.adjoint()
        # TODO(@refraction-ray) omit normalization arg for now
        n = len(ket.out_edges)
        occupied = set()
        nodes = list(ket.nodes) + list(bra.nodes)
        # TODO(@refraction-ray): is the order guaranteed or affect some types of contractor?
        for op, index in ops:
            if not isinstance(op, tn.Node):
                # op is only a matrix
                op = backend.reshaped(op, dim)
                op = gates.Gate(op)
            if isinstance(index, int):
                index = [index]
            noe = len(index)
            for j, e in enumerate(index):
                if e in occupied:
                    raise ValueError("Cannot measure two operators in one index")
                bra.in_edges[e] ^ op.get_edge(j)
                ket.out_edges[e] ^ op.get_edge(j + noe)
                occupied.add(e)
            nodes.append(op)
        for j in range(n):
            if j not in occupied:  # edge1[j].is_dangling invalid here!
                ket.out_edges[j] ^ bra.in_edges[j]
        # self._nodes = nodes1
        num = contractor(nodes).tensor
        return num

    else:
        # ket is the tensor
        if conj is True:
            bra = backend.conj(bra)
        ket = backend.reshape(ket, [-1])
        ket = backend.reshaped(ket, dim)
        bra = backend.reshaped(bra, dim)
        n = len(backend.shape_tuple(ket))
        ket = Gate(ket)
        bra = Gate(bra)
        occupied = set()
        nodes = [ket, bra]
        if normalization is True:
            normket = backend.norm(ket.tensor)
            normbra = backend.norm(bra.tensor)
        for op, index in ops:
            if not isinstance(op, tn.Node):
                # op is only a matrix
                op = backend.reshaped(op, dim)
                op = gates.Gate(op)
            if isinstance(index, int):
                index = [index]
            noe = len(index)
            for j, e in enumerate(index):
                if e in occupied:
                    raise ValueError("Cannot measure two operators in one index")
                bra[e] ^ op.get_edge(j)
                ket[e] ^ op.get_edge(j + noe)
                occupied.add(e)
            nodes.append(op)
        for j in range(n):
            if j not in occupied:  # edge1[j].is_dangling invalid here!
                ket[j] ^ bra[j]
        # self._nodes = nodes1
        num = contractor(nodes).tensor
        if normalization is True:
            den = normket * normbra
        else:
            den = 1.0
        return num / den



"""
Classical shadows functions
"""

from typing import Any, Union, Optional, Sequence, Tuple, List
from string import ascii_letters as ABC

import numpy as np

from .cons import backend, dtypestr, rdtypestr
from .circuit import Circuit

Tensor = Any


def shadow_bound(
    observables: Union[Tensor, Sequence[int]], epsilon: float, delta: float = 0.01
) -> Tuple[int, int]:
    r"""Calculate the shadow bound of the Pauli observables, please refer to the Theorem S1 and Lemma S3 in
    Huang, H.-Y., R. Kueng, and J. Preskill, 2020, Nat. Phys. 16, 1050.

    :param observables: shape = (nq,) or (M, nq), where nq is the number of qubits, M is the number of observables
    :type: Union[Tensor, Sequence[int]]
    :param epsilon: error on the estimator
    :type: float
    :param delta: rate of failure for the bound to hold
    :type: float

    :return Nk: number of snapshots
    :rtype: int
    :return k: number of equal parts to split the shadow snapshot states to compute the median of means.
               k=1 (default) corresponds to simply taking the mean over all shadow snapshot states.
    :rtype: int
    """
    count = np.sign(np.asarray(observables))
    if len(count.shape) == 1:
        count = count[None, :]
    M = count.shape[0]
    k = np.ceil(2 * np.log(2 * M / delta))
    max_length = np.max(np.sum(count, axis=1))
    N = np.ceil((34 / epsilon**2) * 3**max_length)
    return int(N * k), int(k)


def shadow_snapshots(
    psi: Tensor,
    pauli_strings: Tensor,
    status: Optional[Tensor] = None,
    sub: Optional[Sequence[int]] = None,
    measurement_only: bool = False,
) -> Tensor:
    r"""To generate the shadow snapshots from given pauli string observables on psi

    :param psi: shape = (2 ** nq,), where nq is the number of qubits
    :type: Tensor
    :param pauli_strings: shape = (ns, nq), where ns is the number of pauli strings
    :type: Tensor
    :param status: shape = None or (ns, repeat), where repeat is the times to measure on one pauli string
    :type: Optional[Tensor]
    :param sub: qubit indices of subsystem
    :type: Optional[Sequence[int]]
    :param measurement_only: return snapshots (True) or snapshot states (False), default=False
    :type: bool

    :return snapshots: shape = (ns, repeat, nq) if measurement_only=True otherwise (ns, repeat, nq, 2, 2)
    :rtype: Tensor
    """
    pauli_strings = backend.cast(pauli_strings, dtype="int32") - 1
    ns, nq = pauli_strings.shape
    if 2**nq != len(psi):
        raise ValueError(
            f"The number of qubits of psi and pauli_strings should be the same, "
            f"but got {nq} and {int(np.log2(len(psi)))}."
        )
    if status is None:
        status = backend.convert_to_tensor(np.random.rand(ns, 1))
    elif status.shape[0] != ns:
        raise ValueError(f"status.shape[0] should be {ns}, but got {status.shape[0]}.")
    status = backend.cast(status, dtype=rdtypestr)
    repeat = status.shape[1]

    angles = backend.cast(
        backend.convert_to_tensor(
            [
                [-np.pi / 2, np.pi / 4, 0],
                [np.pi / 4, np.pi / 2, 0],
                [0, 0, 0],
            ]
        ),
        dtype=rdtypestr,
    )  # (3, 3)

    def proj_measure(pauli_string: Tensor, st: Tensor) -> Tensor:
        c_ = Circuit(nq, inputs=psi)
        for i in range(nq):
            c_.r(  # type: ignore
                i,
                theta=backend.gather1d(
                    backend.gather1d(angles, backend.gather1d(pauli_string, i)), 0
                ),
                alpha=backend.gather1d(
                    backend.gather1d(angles, backend.gather1d(pauli_string, i)), 1
                ),
                phi=backend.gather1d(
                    backend.gather1d(angles, backend.gather1d(pauli_string, i)), 2
                ),
            )
        return c_.sample(batch=repeat, format="sample_bin", allow_state=True, status=st)

    vpm = backend.vmap(proj_measure, vectorized_argnums=(0, 1))
    snapshots = vpm(pauli_strings, status)  # (ns, repeat, nq)
    if measurement_only:
        return snapshots if sub is None else slice_sub(snapshots, sub)
    else:
        return local_snapshot_states(snapshots, pauli_strings + 1, sub)


def local_snapshot_states(
    snapshots: Tensor, pauli_strings: Tensor, sub: Optional[Sequence[int]] = None
) -> Tensor:
    r"""To generate the local snapshots states from snapshots and pauli strings

    :param snapshots: shape = (ns, repeat, nq)
    :type: Tensor
    :param pauli_strings: shape = (ns, nq) or (ns, repeat, nq)
    :type: Tensor
    :param sub: qubit indices of subsystem
    :type: Optional[Sequence[int]]

    :return lss_states: shape = (ns, repeat, nq, 2, 2)
    :rtype: Tensor
    """
    pauli_strings = backend.cast(pauli_strings, dtype="int32") - 1
    if len(pauli_strings.shape) < len(snapshots.shape):
        pauli_strings = backend.tile(
            pauli_strings[:, None, :], (1, snapshots.shape[1], 1)
        )  # (ns, repeat, nq)

    X_dm = backend.cast(
        backend.convert_to_tensor([[[1, 1], [1, 1]], [[1, -1], [-1, 1]]]) / 2,
        dtype=dtypestr,
    )
    Y_dm = backend.cast(
        backend.convert_to_tensor(
            np.array([[[1, -1j], [1j, 1]], [[1, 1j], [-1j, 1]]]) / 2
        ),
        dtype=dtypestr,
    )
    Z_dm = backend.cast(
        backend.convert_to_tensor([[[1, 0], [0, 0]], [[0, 0], [0, 1]]]), dtype=dtypestr
    )
    pauli_dm = backend.stack((X_dm, Y_dm, Z_dm), axis=0)  # (3, 2, 2, 2)

    def dm(p: Tensor, s: Tensor) -> Tensor:
        return backend.gather1d(backend.gather1d(pauli_dm, p), s)

    v = backend.vmap(dm, vectorized_argnums=(0, 1))
    vv = backend.vmap(v, vectorized_argnums=(0, 1))
    vvv = backend.vmap(vv, vectorized_argnums=(0, 1))

    lss_states = vvv(pauli_strings, snapshots)
    if sub is not None:
        lss_states = slice_sub(lss_states, sub)
    return 3 * lss_states - backend.eye(2)[None, None, None, :, :]


def global_shadow_state(
    snapshots: Tensor,
    pauli_strings: Optional[Tensor] = None,
    sub: Optional[Sequence[int]] = None,
) -> Tensor:
    r"""To generate the global shadow state from local snapshot states or snapshots and pauli strings

    :param snapshots: shape = (ns, repeat, nq, 2, 2) or (ns, repeat, nq)
    :type: Tensor
    :param pauli_strings: shape = None or (ns, nq) or (ns, repeat, nq)
    :type: Optional[Tensor]
    :param sub: qubit indices of subsystem
    :type: Optional[Sequence[int]]

    :return gsdw_state: shape = (2 ** nq, 2 ** nq)
    :rtype: Tensor
    """
    if pauli_strings is not None:
        if len(snapshots.shape) != 3:
            raise ValueError(
                f"snapshots should be 3-d if pauli_strings is not None, got {len(snapshots.shape)}-d instead."
            )
        lss_states = local_snapshot_states(
            snapshots, pauli_strings, sub
        )  # (ns, repeat, nq_sub, 2, 2)
    else:
        if sub is not None:
            lss_states = slice_sub(snapshots, sub)
        else:
            lss_states = snapshots  # (ns, repeat, nq, 2, 2)

    nq = lss_states.shape[2]

    def tensor_prod(dms: Tensor) -> Tensor:
        res = backend.gather1d(dms, 0)
        for i in range(1, nq):
            res = backend.kron(res, backend.gather1d(dms, i))
        return res

    v = backend.vmap(tensor_prod, vectorized_argnums=0)
    vv = backend.vmap(v, vectorized_argnums=0)
    gss_states = vv(lss_states)
    return backend.mean(gss_states, axis=(0, 1))


def expectation_ps_shadow(
    snapshots: Tensor,
    pauli_strings: Optional[Tensor] = None,
    x: Optional[Sequence[int]] = None,
    y: Optional[Sequence[int]] = None,
    z: Optional[Sequence[int]] = None,
    ps: Optional[Sequence[int]] = None,
    k: int = 1,
) -> List[Tensor]:
    r"""To calculate the expectation value of an observable on shadow snapshot states

    :param snapshots: shape = (ns, repeat, nq, 2, 2) or (ns, repeat, nq)
    :type: Tensor
    :param pauli_strings: shape = None or (ns, nq) or (ns, repeat, nq)
    :type: Optional[Tensor]
    :param x: sites to apply X gate, defaults to None
    :type: Optional[Sequence[int]]
    :param y: sites to apply Y gate, defaults to None
    :type: Optional[Sequence[int]]
    :param z: sites to apply Z gate, defaults to None
    :type: Optional[Sequence[int]]
    :param ps: or one can apply a ps structures instead of x, y, z, e.g. [1, 1, 0, 2, 3, 0] for X_0X_1Y_3Z_4
               defaults to None, ps can overwrite x, y and z
    :type: Optional[Sequence[int]]
    :param k: Number of equal parts to split the shadow snapshot states to compute the median of means.
              k=1 (default) corresponds to simply taking the mean over all shadow snapshot states.
    :type: int

    :return expectation values: shape = (k,)
    :rtype: List[Tensor]
    """
    if pauli_strings is not None:
        if len(snapshots.shape) != 3:
            raise ValueError(
                f"snapshots should be 3-d if pauli_strings is not None, got {len(snapshots.shape)}-d instead."
            )
        lss_states = local_snapshot_states(
            snapshots, pauli_strings
        )  # (ns, repeat, nq, 2, 2)
    else:
        lss_states = snapshots  # (ns, repeat, nq, 2, 2)
    ns, repeat, nq, _, _ = lss_states.shape
    ns *= repeat
    ss_states = backend.reshape(lss_states, (ns, nq, 2, 2))

    if ps is None:
        ps = [0] * nq
        if x is not None:
            for i in x:
                ps[i] = 1
        if y is not None:
            for i in y:
                ps[i] = 2
        if z is not None:
            for i in z:
                ps[i] = 3
    elif len(ps) != nq:
        raise ValueError(
            f"The number of qubits of the shadow state is {nq}, but got a {len(ps)}-qubit pauli string observable."
        )
    ps = backend.cast(backend.convert_to_tensor(ps), dtype="int32")  # (nq,)

    paulis = backend.convert_to_tensor(
        backend.cast(
            np.array(
                [
                    [[1, 0], [0, 1]],
                    [[0, 1], [1, 0]],
                    [[0, -1j], [1j, 0]],
                    [[1, 0], [0, -1]],
                ]
            ),
            dtype=dtypestr,
        )
    )  # (4, 2, 2)

    def trace_paulis_prod(dm: Tensor, idx: Tensor) -> Tensor:
        return backend.real(backend.trace(backend.gather1d(paulis, idx) @ dm))

    v = backend.vmap(trace_paulis_prod, vectorized_argnums=(0, 1))  # (nq,)

    def prod(dm: Tensor) -> Tensor:
        return backend.shape_prod(v(dm, ps))

    vv = backend.vmap(prod, vectorized_argnums=0)  # (ns,)

    batch = int(np.ceil(ns / k))
    return [backend.mean(vv(ss_states[i : i + batch])) for i in range(0, ns, batch)]


def entropy_shadow(
    snapshots: Tensor,
    pauli_strings: Optional[Tensor] = None,
    sub: Optional[Sequence[int]] = None,
    alpha: int = 2,
) -> Tensor:
    r"""To calculate the Renyi entropy of a subsystem from shadow state or shadow snapshot states

    :param snapshots: shape = (ns, repeat, nq, 2, 2) or (ns, repeat, nq)
    :type: Tensor
    :param pauli_strings: shape = None or (ns, nq) or (ns, repeat, nq)
    :type: Optional[Tensor]
    :param sub: qubit indices of subsystem
    :type: Optional[Sequence[int]]
    :param alpha: order of the Renyi entropy, alpha=1 corresponds to the von Neumann entropy
    :type: int

    :return Renyi entropy: shape = ()
    :rtype: Tensor
    """
    if alpha <= 0:
        raise ValueError("Alpha should not be less than 1!")

    sdw_rdm = global_shadow_state(snapshots, pauli_strings, sub)  # (2 ** nq, 2 ** nq)

    evs = backend.relu(backend.real(backend.eigvalsh(sdw_rdm)))
    evs /= backend.sum(evs)
    if alpha == 1:
        return -backend.sum(evs * backend.log(evs + 1e-12))
    else:
        return backend.log(backend.sum(backend.power(evs, alpha))) / (1 - alpha)


def renyi_entropy_2(snapshots: Tensor, sub: Optional[Sequence[int]] = None) -> Tensor:
    r"""To calculate the second order Renyi entropy of a subsystem from snapshot, please refer to
    Brydges, T. et al. Science 364, 260-263 (2019). This function is not jitable.

    :param snapshots: shape = (ns, repeat, nq)
    :type: Tensor
    :param sub: qubit indices of subsystem
    :type: Optional[Sequence[int]]

    :return second order Renyi entropy: shape = ()
    :rtype: Tensor
    """
    if sub is not None:
        snapshots = slice_sub(snapshots, sub)
    snapshots = backend.cast(snapshots, dtype="int32")
    ns, repeat, nq = snapshots.shape

    count = {}
    for i, ss in enumerate(snapshots):
        for s in ss:
            s = tuple(backend.numpy(s))
            if s not in count:
                count[s] = np.zeros(ns)
            count[s][i] += 1

    tr = 0.0
    for x in count:
        for y in count:
            h = np.sum((np.asarray(x) + np.asarray(y)) % 2)
            pp_mean = np.mean(count[x] * count[y]) / repeat / repeat
            tr += pp_mean * (-2.0) ** (-h)
    return -np.log(tr * 2**nq)


def global_shadow_state1(
    snapshots: Tensor,
    pauli_strings: Optional[Tensor] = None,
    sub: Optional[Sequence[int]] = None,
) -> Tensor:
    r"""To generate the global snapshots states from local snapshot states or snapshots and pauli strings

    :param snapshots: shape = (ns, repeat, nq, 2, 2) or (ns, repeat, nq)
    :type: Tensor
    :param pauli_strings: shape = None or (ns, nq) or (ns, repeat, nq)
    :type: Optional[Tensor]
    :param sub: qubit indices of subsystem
    :type: Optional[Sequence[int]]

    :return gsdw_state: shape = (2 ** nq, 2 ** nq)
    :rtype: Tensor
    """
    if pauli_strings is not None:
        if len(snapshots.shape) != 3:
            raise ValueError(
                f"snapshots should be 3-d if pauli_strings is not None, got {len(snapshots.shape)}-d instead."
            )
        lss_states = local_snapshot_states(
            snapshots, pauli_strings, sub
        )  # (ns, repeat, nq_sub, 2, 2)
    else:
        if sub is not None:
            lss_states = slice_sub(snapshots, sub)
        else:
            lss_states = snapshots  # (ns, repeat, nq, 2, 2)
    lss_states = backend.transpose(
        lss_states, (2, 0, 1, 3, 4)
    )  # (nq, ns, repeat, 2, 2)
    nq, ns, repeat, _, _ = lss_states.shape

    old_indices = [f"ab{ABC[2 + 2 * i: 4 + 2 * i]}" for i in range(nq)]
    new_indices = f"ab{ABC[2:2 * nq + 2:2]}{ABC[3:2 * nq + 2:2]}"

    gss_states = backend.reshape(
        backend.einsum(
            f'{",".join(old_indices)}->{new_indices}', *lss_states, optimize=True
        ),
        (ns, repeat, 2**nq, 2**nq),
    )
    return backend.mean(gss_states, axis=(0, 1))


def global_shadow_state2(
    snapshots: Tensor,
    pauli_strings: Optional[Tensor] = None,
    sub: Optional[Sequence[int]] = None,
) -> Tensor:
    r"""To generate the global snapshots states from local snapshot states or snapshots and pauli strings

    :param snapshots: shape = (ns, repeat, nq, 2, 2) or (ns, repeat, nq)
    :type: Tensor
    :param pauli_strings: shape = None or (ns, nq) or (ns, repeat, nq)
    :type: Optional[Tensor]
    :param sub: qubit indices of subsystem
    :type: Optional[Sequence[int]]

    :return gsdw_state: shape = (2 ** nq, 2 ** nq)
    :rtype: Tensor
    """
    if pauli_strings is not None:
        if len(snapshots.shape) != 3:
            raise ValueError(
                f"snapshots should be 3-d if pauli_strings is not None, got {len(snapshots.shape)}-d instead."
            )
        lss_states = local_snapshot_states(
            snapshots, pauli_strings, sub
        )  # (ns, repeat, nq_sub, 2, 2)
    else:
        if sub is not None:
            lss_states = slice_sub(snapshots, sub)
        else:
            lss_states = snapshots  # (ns, repeat, nq, 2, 2)
    nq = lss_states.shape[2]

    old_indices = [f"{ABC[2 * i: 2 + 2 * i]}" for i in range(nq)]
    new_indices = f"{ABC[0:2 * nq:2]}{ABC[1:2 * nq:2]}"

    def tensor_prod(dms: Tensor) -> Tensor:
        return backend.reshape(
            backend.einsum(
                f'{",".join(old_indices)}->{new_indices}', *dms, optimize=True
            ),
            (2**nq, 2**nq),
        )

    v = backend.vmap(tensor_prod, vectorized_argnums=0)
    vv = backend.vmap(v, vectorized_argnums=0)
    gss_states = vv(lss_states)
    return backend.mean(gss_states, axis=(0, 1))


def slice_sub(entirety: Tensor, sub: Sequence[int]) -> Tensor:
    r"""To slice off the subsystem

    :param entirety: shape = (ns, repeat, nq, 2, 2) or (ns, repeat, nq)
    :type: Tensor
    :param sub: qubit indices of subsystem
    :type: Sequence[int]

    :return subsystem: shape = (ns, repeat, nq_sub, 2, 2)
    :rtype: Tensor
    """
    if len(entirety.shape) < 3:
        entirety = entirety[:, None, :]

    def slc(x: Tensor, idx: Tensor) -> Tensor:
        return backend.gather1d(x, idx)

    v = backend.vmap(slc, vectorized_argnums=(1,))
    vv = backend.vmap(v, vectorized_argnums=(0,))
    vvv = backend.vmap(vv, vectorized_argnums=(0,))
    return vvv(entirety, backend.convert_to_tensor(sub))



"""
CuPy backend. Not in the tensornetwork package and highly experimental.
"""

# pylint: disable=invalid-name

import logging
import warnings
from typing import Any, Callable, Optional, Sequence, Tuple, Union

import numpy as np

try:
    from numpy import ComplexWarning
except ImportError:  # np2.0 compatibility
    from numpy.exceptions import ComplexWarning  # type: ignore

import scipy

try:  # old version tn compatiblity
    from tensornetwork.backends import base_backend

    tnbackend = base_backend.BaseBackend

except ImportError:
    from tensornetwork.backends import abstract_backend

    tnbackend = abstract_backend.AbstractBackend


from .abstract_backend import ExtendedBackend

logger = logging.getLogger(__name__)

dtypestr: str
Tensor = Any

cp: Any
cpx: Any


class CuPyBackend(tnbackend, ExtendedBackend):  # type: ignore
    def __init__(self) -> None:
        super().__init__()
        try:
            import cupy
            import cupyx
        except ImportError:
            raise ImportError(
                "CuPy not installed, please switch to a different "
                "backend or install CuPy."
            )
        global cp
        global cpx
        cp = cupy
        cpx = cupyx
        self.name = "cupy"

    def convert_to_tensor(self, a: Tensor, dtype: Optional[str] = None) -> Tensor:
        if not isinstance(a, cp.ndarray) and not cp.isscalar(a):
            a = cp.array(a)
        a = cp.asarray(a)
        if dtype is not None:
            a = self.cast(a, dtype)
        return a

    def sum(
        self: Any,
        a: Tensor,
        axis: Optional[Sequence[int]] = None,
        keepdims: bool = False,
    ) -> Tensor:
        return cp.sum(a, axis=axis, keepdims=keepdims)

    def conj(self, tensor: Tensor) -> Tensor:
        return tensor.conj()

    def sign(self, tensor: Tensor) -> Tensor:
        return cp.sign(tensor)

    def multiply(self, tensor1: Tensor, tensor2: Tensor) -> Tensor:
        return tensor1 * tensor2

    def norm(self, tensor: Tensor) -> Tensor:
        return cp.linalg.norm(tensor)

    def shape_tuple(self, tensor: Tensor) -> Tuple[int]:
        return tensor.shape  # type:ignore

    def tensordot(
        self, a: Tensor, b: Tensor, axes: Union[int, Sequence[Sequence[int]]]
    ) -> Tensor:
        return cp.tensordot(a, b, axes)

    def outer_product(self, tensor1: Tensor, tensor2: Tensor) -> Tensor:
        return cp.tensordot(tensor1, tensor2, 0)

    def transpose(self, tensor: Tensor, perm: Optional[Sequence[int]] = None) -> Tensor:
        return cp.transpose(tensor, perm)

    def reshape(self, tensor: Tensor, shape: Tensor) -> Tensor:
        return cp.reshape(tensor, np.asarray(shape).astype(np.int32))

    def eye(
        self, N: int, dtype: Optional[str] = None, M: Optional[int] = None
    ) -> Tensor:
        if dtype is None:
            dtype = dtypestr
        return cp.eye(N, M=M, dtype=dtype)

    def ones(self, shape: Sequence[int], dtype: Optional[str] = None) -> Tensor:
        if dtype is None:
            dtype = dtypestr
        return cp.ones(shape, dtype=dtype)

    def zeros(self, shape: Sequence[int], dtype: Optional[str] = None) -> Tensor:
        if dtype is None:
            dtype = dtypestr
        return cp.zeros(shape, dtype=dtype)

    def copy(self, a: Tensor) -> Tensor:
        return a.copy()

    def expm(self, a: Tensor) -> Tensor:
        return self.convert_to_tensor(scipy.linalg.expm(self.numpy(a)))

    def abs(self, a: Tensor) -> Tensor:
        return cp.abs(a)

    def sin(self, a: Tensor) -> Tensor:
        return cp.sin(a)

    def cos(self, a: Tensor) -> Tensor:
        return cp.cos(a)

    # acos acosh asin asinh atan atan2 atanh cosh (cos) tan tanh sinh (sin)
    def acos(self, a: Tensor) -> Tensor:
        return cp.arccos(a)

    def acosh(self, a: Tensor) -> Tensor:
        return cp.arccosh(a)

    def asin(self, a: Tensor) -> Tensor:
        return cp.arcsin(a)

    def asinh(self, a: Tensor) -> Tensor:
        return cp.arcsinh(a)

    def atan(self, a: Tensor) -> Tensor:
        return cp.arctan(a)

    def atan2(self, y: Tensor, x: Tensor) -> Tensor:
        return cp.arctan2(y, x)

    def atanh(self, a: Tensor) -> Tensor:
        return cp.arctanh(a)

    def cosh(self, a: Tensor) -> Tensor:
        return cp.cosh(a)

    def tan(self, a: Tensor) -> Tensor:
        return cp.tan(a)

    def tanh(self, a: Tensor) -> Tensor:
        return cp.tanh(a)

    def sinh(self, a: Tensor) -> Tensor:
        return cp.sinh(a)

    def size(self, a: Tensor) -> Tensor:
        return a.size

    def eigvalsh(self, a: Tensor) -> Tensor:
        return cp.linalg.eigvalsh(a)

    def kron(self, a: Tensor, b: Tensor) -> Tensor:
        return cp.kron(a, b)

    def dtype(self, a: Tensor) -> str:
        return a.dtype.__str__()  # type: ignore

    def numpy(self, a: Tensor) -> Tensor:
        if isinstance(a, cp.ndarray):
            return a.get()
        else:
            return np.array(a)

    def i(self, dtype: Any = None) -> Tensor:
        if not dtype:
            dtype = npdtype  # type: ignore
        if isinstance(dtype, str):
            dtype = getattr(np, dtype)
        return cp.array(1j, dtype=dtype)

    def stack(self, a: Sequence[Tensor], axis: int = 0) -> Tensor:
        return cp.stack(a, axis=axis)

    def concat(self, a: Sequence[Tensor], axis: int = 0) -> Tensor:
        return cp.concatenate(a, axis=axis)

    def tile(self, a: Tensor, rep: Tensor) -> Tensor:
        return cp.tile(a, rep)

    def mean(
        self,
        a: Tensor,
        axis: Optional[Sequence[int]] = None,
        keepdims: bool = False,
    ) -> Tensor:
        return cp.mean(a, axis=axis, keepdims=keepdims)

    def std(
        self, a: Tensor, axis: Optional[Sequence[int]] = None, keepdims: bool = False
    ) -> Tensor:
        return cp.std(a, axis=axis, keepdims=keepdims)

    def unique_with_counts(self, a: Tensor, **kws: Any) -> Tuple[Tensor, Tensor]:
        return cp.unique(a, return_counts=True)  # type: ignore

    def min(self, a: Tensor, axis: Optional[int] = None) -> Tensor:
        return cp.min(a, axis=axis)

    def max(self, a: Tensor, axis: Optional[int] = None) -> Tensor:
        return cp.max(a, axis=axis)

    def argmax(self, a: Tensor, axis: int = 0) -> Tensor:
        return cp.argmax(a, axis=axis)

    def argmin(self, a: Tensor, axis: int = 0) -> Tensor:
        return cp.argmin(a, axis=axis)

    def sigmoid(self, a: Tensor) -> Tensor:
        return cpx.scipy.special.expit(a)

    def relu(self, a: Tensor) -> Tensor:
        return (abs(a) + a) / 2
        # this impl seems to be the fastest
        # see https://stackoverflow.com/questions/32109319/how-to-implement-the-relu-function-in-numpy

    def softmax(self, a: Sequence[Tensor], axis: Optional[int] = None) -> Tensor:
        return cpx.scipy.special.softmax(a, axis=axis)

    def onehot(self, a: Tensor, num: int) -> Tensor:
        res = cp.eye(num)[a.reshape([-1])]
        return res.reshape(list(a.shape) + [num])
        # https://stackoverflow.com/questions/38592324/one-hot-encoding-using-numpy

    def cumsum(self, a: Tensor, axis: Optional[int] = None) -> Tensor:
        return cp.cumsum(a, axis)

    def is_tensor(self, a: Any) -> bool:
        if isinstance(a, cp.ndarray):
            return True
        return False

    def real(self, a: Tensor) -> Tensor:
        return cp.real(a)

    def imag(self, a: Tensor) -> Tensor:
        return cp.imag(a)

    def cast(self, a: Tensor, dtype: str) -> Tensor:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ComplexWarning)
            if isinstance(dtype, str):
                return a.astype(getattr(np, dtype))
            return a.astype(dtype)

    def arange(self, start: int, stop: Optional[int] = None, step: int = 1) -> Tensor:
        if stop is None:
            return cp.arange(start=0, stop=start, step=step)
        return cp.arange(start=start, stop=stop, step=step)

    def mod(self, x: Tensor, y: Tensor) -> Tensor:
        return cp.mod(x, y)

    def right_shift(self, x: Tensor, y: Tensor) -> Tensor:
        return cp.right_shift(x, y)

    def left_shift(self, x: Tensor, y: Tensor) -> Tensor:
        return cp.left_shift(x, y)

    def solve(self, A: Tensor, b: Tensor, assume_a: str = "gen") -> Tensor:  # type: ignore
        raise NotImplementedError

    def searchsorted(self, a: Tensor, v: Tensor, side: str = "left") -> Tensor:
        return cp.searchsorted(a, v, side=side)

    def set_random_state(
        self, seed: Optional[int] = None, get_only: bool = False
    ) -> Any:
        g = cp.random.default_rng(seed)  # None auto supported
        if get_only is False:
            self.g = g
        return g

    def stateful_randn(
        self,
        g: "cp.random.Generator",
        shape: Union[int, Sequence[int]] = 1,
        mean: float = 0,
        stddev: float = 1,
        dtype: str = "32",
    ) -> Tensor:
        if isinstance(dtype, str):
            dtype = dtype[-2:]
        if isinstance(shape, int):
            shape = (shape,)
        r = g.normal(loc=mean, scale=stddev, size=shape)
        if dtype == "32":
            r = r.astype(np.float32)
        elif dtype == "64":
            r = r.astype(np.float64)
        elif not isinstance(dtype, str):
            r = r.astype(dtype)
        else:
            raise ValueError("unspported `dtype` %s" % dtype)
        return r

    def stateful_randu(
        self,
        g: "cp.random.Generator",
        shape: Union[int, Sequence[int]] = 1,
        low: float = 0,
        high: float = 1,
        dtype: str = "32",
    ) -> Tensor:
        if isinstance(dtype, str):
            dtype = dtype[-2:]
        if isinstance(shape, int):
            shape = (shape,)
        r = g.random(shape) * (high - low) + low
        if dtype == "32":
            r = r.astype(np.float32)
        elif dtype == "64":
            r = r.astype(np.float64)
        elif not isinstance(dtype, str):
            r = r.astype(dtype)
        else:
            raise ValueError("unspported `dtype` %s" % dtype)
        return r

    def stateful_randc(
        self,
        g: "cp.random.Generator",
        a: Union[int, Sequence[int], Tensor],
        shape: Union[int, Sequence[int]],
        p: Optional[Union[Sequence[float], Tensor]] = None,
    ) -> Tensor:
        if isinstance(shape, int):
            shape = (shape,)
        return g.choice(a, size=shape, replace=True, p=p)

    def scatter(self, operand: Tensor, indices: Tensor, updates: Tensor) -> Tensor:
        operand_new = cp.copy(operand)
        operand_new[tuple([indices[:, i] for i in range(indices.shape[1])])] = updates
        return operand_new

    def coo_sparse_matrix(
        self, indices: Tensor, values: Tensor, shape: Tensor
    ) -> Tensor:
        values = self.convert_to_tensor(values)
        indices = self.convert_to_tensor(indices).T
        return cp.sparse.coo_matrix((values, indices), shape=shape)

    def sparse_dense_matmul(
        self,
        sp_a: Tensor,
        b: Tensor,
    ) -> Tensor:
        return sp_a @ b

    def to_dense(self, sp_a: Tensor) -> Tensor:
        return sp_a.todense()

    def is_sparse(self, a: Tensor) -> bool:
        return cpx.scipy.sparse.issparse(a)  # type: ignore

    def cond(
        self,
        pred: bool,
        true_fun: Callable[[], Tensor],
        false_fun: Callable[[], Tensor],
    ) -> Tensor:
        if pred:
            return true_fun()
        return false_fun()

    def switch(self, index: Tensor, branches: Sequence[Callable[[], Tensor]]) -> Tensor:
        return branches[index]()

    def device(self, a: Tensor) -> str:
        return self._dev2str(a.device)

    def device_move(self, a: Tensor, dev: Any) -> Tensor:
        if isinstance(dev, str):
            dev = self._str2dev(dev)
        with dev:
            return cp.asarray(a)

    def _dev2str(self, dev: Any) -> str:
        return f"gpu:{dev.id}"

    def _str2dev(self, str_: str) -> Any:
        if str_ == "cpu":
            raise ValueError("CuPy backend only support GPU device")
        else:
            return cp.cuda.Device(int(str_.split(":")[-1]))

    def stop_gradient(self, a: Tensor) -> Tensor:
        raise NotImplementedError("CuPy backend doesn't support AD")

    def grad(
        self,
        f: Callable[..., Any],
        argnums: Union[int, Sequence[int]] = 0,
        has_aux: bool = False,
    ) -> Callable[..., Any]:
        raise NotImplementedError("CuPy backend doesn't support AD")

    def value_and_grad(
        self,
        f: Callable[..., Any],
        argnums: Union[int, Sequence[int]] = 0,
        has_aux: bool = False,
    ) -> Callable[..., Tuple[Any, Any]]:
        raise NotImplementedError("CuPy backend doesn't support AD")

    def jit(
        self,
        f: Callable[..., Any],
        static_argnums: Optional[Union[int, Sequence[int]]] = None,
        jit_compile: Optional[bool] = None,
        **kws: Any,
    ) -> Callable[..., Any]:
        logger.warning("CuPy backend has no jit interface, just do nothing")
        return f
        # raise NotImplementedError("numpy backend doesn't support jit compiling")

    def vmap(
        self, f: Callable[..., Any], vectorized_argnums: Union[int, Sequence[int]] = 0
    ) -> Any:
        logger.warning(
            "CuPy backend has no intrinsic vmap like interface"
            ", use vectorize instead (plain for loop)"
        )
        if isinstance(vectorized_argnums, int):
            vectorized_argnums = (vectorized_argnums,)

        def wrapper(*args: Any, **kws: Any) -> Tensor:
            results = []
            for barg in zip(*[args[i] for i in vectorized_argnums]):  # type: ignore
                narg = []
                j = 0
                for k in range(len(args)):
                    if k in vectorized_argnums:  # type: ignore
                        narg.append(barg[j])
                        j += 1
                    else:
                        narg.append(args[k])
                results.append(f(*narg, **kws))
            return cp.array(results)

        return wrapper

    def vectorized_value_and_grad(
        self,
        f: Callable[..., Any],
        argnums: Union[int, Sequence[int]] = 0,
        vectorized_argnums: Union[int, Sequence[int]] = 0,
        has_aux: bool = False,
    ) -> Callable[..., Tuple[Any, Any]]:
        raise NotImplementedError("CuPy backend doesn't support AD")

    vvag = vectorized_value_and_grad

    def vjp(
        self,
        f: Callable[..., Any],
        inputs: Union[Tensor, Sequence[Tensor]],
        v: Union[Tensor, Sequence[Tensor]],
    ) -> Tuple[Union[Tensor, Sequence[Tensor]], Union[Tensor, Sequence[Tensor]]]:
        raise NotImplementedError("CuPy backend doesn't support AD")

    def jvp(
        self,
        f: Callable[..., Any],
        inputs: Union[Tensor, Sequence[Tensor]],
        v: Union[Tensor, Sequence[Tensor]],
    ) -> Tuple[Union[Tensor, Sequence[Tensor]], Union[Tensor, Sequence[Tensor]]]:
        raise NotImplementedError("CuPy backend doesn't support AD")



"""
Quantum circuit: state simulator for **qudits** (d-level systems).

This module provides a high-level `QuditCircuit` API that mirrors `tensorcircuit.circuit.Circuit`
but targets qudits with dimension `3 <= d <= 36`.
For string-encoded samples/counts, digits use `0-9A-Z` where `A=10, ..., Z=35`.

.. note::
   For qubits (`d=2`) please use :class:`tensorcircuit.circuit.Circuit`.

"""

from functools import partial
from typing import Any, Dict, List, Optional, Tuple, Sequence, Union, Literal

import numpy as np
import tensornetwork as tn

from .gates import Gate
from .utils import arg_alias
from .basecircuit import BaseCircuit
from .circuit import Circuit
from .quantum import QuOperator, QuVector
from .quditgates import SINGLE_BUILDERS, TWO_BUILDERS


Tensor = Any
SAMPLE_FORMAT = Literal["sample_bin", "count_dict_bin"]


class QuditCircuit:
    r"""
    The `QuditCircuit` class provides a d-dimensional state-vector simulator and a thin wrapper
    around :class:`tensorcircuit.circuit.Circuit`, exposing a qudit-friendly API and docstrings.

    **Quick example (d=3):**

    >>> c = tc.QuditCircuit(2, dim=3)
    >>> c.h(0)
    >>> c.x(1)
    >>> c.csum(0, 1)
    >>> c.sample(1024, format="count_dict_bin")

    .. note::
       For `3 <= d <= 36`, string samples and count keys use base-`d` characters `0-9A-Z`
       (`A=10, ..., Z=35`).

    :param nqudits: Number of qudits (wires) in the circuit.
    :type nqudits: int
    :param dim: Qudit local dimension `d`. Must satisfy `3 <= d <= 36`.
    :type dim: int
    :param inputs: Optional initial state as a wavefunction.
    :type inputs: Optional[Tensor]
    :param mps_inputs: Optional initial state in MPS/MPO-like form.
    :type mps_inputs: Optional[QuOperator]
    :param split: Internal contraction/splitting configuration passed through to
                  :class:`~tensorcircuit.circuit.Circuit`.
    :type split: Optional[Dict[str, Any]]

    :var is_dm: Whether the simulator is a density-matrix simulator (`False` here).
    :vartype is_dm: bool
    :var dim: Property for the local dimension `d`.
    :vartype dim: int
    :var nqudits: Property for the number of qudits.
    :vartype nqudits: int
    """

    is_dm = False

    def __init__(
        self,
        nqudits: int,
        dim: int,
        inputs: Optional[Tensor] = None,
        mps_inputs: Optional[QuOperator] = None,
        split: Optional[Dict[str, Any]] = None,
    ):
        self._set_dim(dim=dim)
        self._nqudits = nqudits

        self._circ = Circuit(
            nqubits=nqudits,
            dim=dim,
            inputs=inputs,
            mps_inputs=mps_inputs,
            split=split,
        )
        self._omega = np.exp(2j * np.pi / self._d)
        self.circuit_param = self._circ.circuit_param

    def _set_dim(self, dim: int) -> None:
        if not isinstance(dim, int) or dim <= 2:
            raise ValueError(
                f"QuditCircuit is only for qudits (dim>=3). "
                f"You passed dim={dim}. For qubits, please use `Circuit` instead."
            )
        # Require integer d>=2; current string-encoded IO supports d<=36 (0-9A-Z digits).
        if dim > 36:
            raise NotImplementedError(
                "The Qudit interface is only supported for dimension < 36 now."
            )
        self._d = dim

    @property
    def dim(self) -> int:
        """dimension of the qudit circuit"""
        return self._d

    @property
    def nqudits(self) -> int:
        """qudit number of the circuit"""
        return self._nqudits

    def _apply_gate(self, *indices: int, name: str, **kwargs: Any) -> None:
        """
        Apply a single- or two-qudit unitary by name.

        The gate matrix is looked up by name in either :data:`SINGLE_BUILDERS` (single-qudit)
        or :data:`TWO_BUILDERS` (two-qudit). The registered builder is called with `(d, omega, **kwargs)`
        to produce the unitary, which is then applied at the given indices.

        :param indices: Qudit indices the gate acts on. One index -> single-qudit gate; two indices -> two-qudit gate.
        :type indices: int
        :param name: Gate name registered in the corresponding builder set.
        :type name: str
        :param kwargs: Extra parameters forwarded to the builder (matched by keyword).
        :type kwargs: Any
        :raises ValueError: If the name is unknown or the arity does not match the number of indices.
        """
        if len(indices) == 1 and name in SINGLE_BUILDERS:
            sig, builder = SINGLE_BUILDERS[name]
            extras = tuple(kwargs.get(k) for k in sig if k != "none")
            builder_kwargs = {k: v for k, v in zip(sig, extras)}
            mat = builder(self._d, self._omega, **builder_kwargs)
            self._circ.unitary(*indices, unitary=mat, name=name, dim=self._d)  # type: ignore
        elif len(indices) == 2 and name in TWO_BUILDERS:
            sig, builder = TWO_BUILDERS[name]
            extras = tuple(kwargs.get(k) for k in sig if k != "none")
            builder_kwargs = {k: v for k, v in zip(sig, extras)}
            mat = builder(self._d, self._omega, **builder_kwargs)
            self._circ.unitary(  # type: ignore
                *indices, unitary=mat, name=name, dim=self._d
            )
        else:
            raise ValueError(f"Unsupported gate/arity: {name} on {len(indices)} qudits")

    def any(self, *indices: int, unitary: Tensor, name: Optional[str] = None) -> None:
        """
        Apply an arbitrary unitary on one or two qudits.

        :param indices: Target qudit indices.
        :type indices: int
        :param unitary: Unitary matrix acting on the specified qudit(s), with shape `(d, d)` or `(d^2, d^2)`.
        :type unitary: Tensor
        :param name: Optional label stored in the circuit history.
        :type name: str
        """
        name = "any" if name is None else name
        self._circ.unitary(*indices, unitary=unitary, name=name, dim=self._d)  # type: ignore

    unitary = any

    def i(self, index: int) -> None:
        """
        Apply the generalized identity gate `I` on the given qudit.

        :param index: Qudit index.
        :type index: int
        """
        self._apply_gate(index, name="I")

    I = i

    def x(self, index: int) -> None:
        """
        Apply the generalized shift gate `X` on the given qudit (adds `+1 mod d`).

        :param index: Qudit index.
        :type index: int
        """
        self._apply_gate(index, name="X")

    X = x

    # def y(self, index: int) -> None:
    #     """
    #     Apply the Y gate on the given qudit index.
    #
    #     :param index: Qudit index to apply the gate on.
    #     :type index: int
    #     """
    #     self._apply_gate(index, name="Y")

    def z(self, index: int) -> None:
        """
        Apply the generalized phase gate `Z` on the given qudit (multiplies by `omega^k`).

        :param index: Qudit index.
        :type index: int
        """
        self._apply_gate(index, name="Z")

    Z = z

    def h(self, index: int) -> None:
        """
        Apply the generalized Hadamard-like gate `H` (DFT on `d` levels) on the given qudit.

        :param index: Qudit index.
        :type index: int
        """
        self._apply_gate(index, name="H")

    H = h

    def u8(
        self, index: int, gamma: float = 2.0, z: float = 1.0, eps: float = 0.0
    ) -> None:
        """
        Apply the parameterized single-qudit gate `U8`.

        :param index: Qudit index.
        :type index: int
        :param gamma: Gate parameter `gamma`.
        :type gamma: float
        :param z: Gate parameter `z`.
        :type z: float
        :param eps: Gate parameter `eps`.
        :type eps: float
        """
        self._apply_gate(index, name="U8", extra=(gamma, z, eps))

    U8 = u8

    def rx(self, index: int, theta: float, j: int = 0, k: int = 1) -> None:
        """
        Single-qudit rotation `RX` on a selected two-level subspace `(j, k)`.

        :param index: Qudit index.
        :type index: int
        :param theta: Rotation angle.
        :type theta: float
        :param j: Source level of the rotation subspace (0-based).
        :type j: int
        :param k: Target level of the rotation subspace (0-based).
        :type k: int
        :raises ValueError: If `j == k` or indices are outside `[0, d-1]`.
        """
        self._apply_gate(index, name="RX", theta=theta, j=j, k=k)

    RX = rx

    def ry(self, index: int, theta: float, j: int = 0, k: int = 1) -> None:
        """
        Single-qudit rotation `RY` on a selected two-level subspace `(j, k)`.

        :param index: Qudit index.
        :type index: int
        :param theta: Rotation angle.
        :type theta: float
        :param j: Source level of the rotation subspace (0-based).
        :type j: int
        :param k: Target level of the rotation subspace (0-based).
        :type k: int
        :raises ValueError: If `j == k` or indices are outside `[0, d-1]`.
        """
        self._apply_gate(index, name="RY", theta=theta, j=j, k=k)

    RY = ry

    def rz(self, index: int, theta: float, j: int = 0) -> None:
        """
        Single-qudit phase rotation `RZ` applied on level `j`.

        :param index: Qudit index.
        :type index: int
        :param theta: Phase rotation angle around Z.
        :type theta: float
        :param j: Level where the phase is applied (0-based).
        :type j: int
        :raises ValueError: If `j` is outside `[0, d-1]`.
        """
        self._apply_gate(index, name="RZ", theta=theta, j=j)

    RZ = rz

    def rxx(
        self,
        *indices: int,
        theta: float,
        j1: int = 0,
        k1: int = 1,
        j2: int = 0,
        k2: int = 1,
    ) -> None:
        """
        Two-qudit interaction `RXX` acting on subspaces `(j1, k1)` and `(j2, k2)`.

        :param indices: Two qudit indices `(q1, q2)`.
        :type indices: int
        :param theta: Interaction angle.
        :type theta: float
        :param j1: Source level of the first qudit subspace.
        :type j1: int
        :param k1: Target level of the first qudit subspace.
        :type k1: int
        :param j2: Source level of the second qudit subspace.
        :type j2: int
        :param k2: Target level of the second qudit subspace.
        :type k2: int
        :raises ValueError: If levels are invalid or the arity is not two.
        """
        self._apply_gate(*indices, name="RXX", theta=theta, j1=j1, k1=k1, j2=j2, k2=k2)

    RXX = rxx

    def rzz(
        self,
        *indices: int,
        theta: float,
        j1: int = 0,
        k1: int = 1,
        j2: int = 0,
        k2: int = 1,
    ) -> None:
        """
        Two-qudit interaction `RZZ` acting on subspaces `(j1, k1)` and `(j2, k2)`.

        :param indices: Two qudit indices `(q1, q2)`.
        :type indices: int
        :param theta: Interaction angle.
        :type theta: float
        :param j1: Source level of the first qudit subspace.
        :type j1: int
        :param k1: Target level of the first qudit subspace.
        :type k1: int
        :param j2: Source level of the second qudit subspace.
        :type j2: int
        :param k2: Target level of the second qudit subspace.
        :type k2: int
        :raises ValueError: If levels are invalid or the arity is not two.
        """
        self._apply_gate(*indices, name="RZZ", theta=theta, j1=j1, k1=k1, j2=j2, k2=k2)

    RZZ = rzz

    def cphase(self, *indices: int, cv: Optional[int] = None) -> None:
        """
        Apply a controlled-phase gate `CPHASE`.

        :param indices: Two qudit indices `(control, target)`.
        :type indices: int
        :param cv: Optional control value. If `None`, defaults to `1`.
        :type cv: Optional[int]
        :raises ValueError: If arity is not two.
        """
        self._apply_gate(*indices, name="CPHASE", cv=cv)

    CPHASE = cphase

    def csum(self, *indices: int, cv: Optional[int] = None) -> None:
        """
        Apply a generalized controlled-sum gate `CSUM` (a.k.a. qudit CNOT).

        :param indices: Two qudit indices `(control, target)`.
        :type indices: int
        :param cv: Optional control value. If `None`, defaults to `1`.
        :type cv: Optional[int]
        :raises ValueError: If arity is not two.
        """
        self._apply_gate(*indices, name="CSUM", cv=cv)

    cnot, CSUM, CNOT = csum, csum, csum

    # Functional
    def wavefunction(self, form: str = "default") -> tn.Node.tensor:
        """
        Compute the output wavefunction from the circuit.

        :param form: The str indicating the form of the output wavefunction.
            "default": [-1], "ket": [-1, 1], "bra": [1, -1]
        :type form: str, optional
        :return: Tensor with the corresponding shape.
        :rtype: Tensor
        """
        return self._circ.wavefunction(form)

    state = wavefunction

    def get_quoperator(self) -> QuOperator:
        """
        Get the ``QuOperator`` MPO like representation of the circuit unitary without contraction.

        :return: ``QuOperator`` object for the circuit unitary (open indices for the input state)
        :rtype: QuOperator
        """
        return self._circ.quoperator()

    quoperator = get_quoperator

    get_circuit_as_quoperator = get_quoperator
    get_state_as_quvector = BaseCircuit.quvector

    def matrix(self) -> Tensor:
        """
        Get the unitary matrix for the circuit irrespective with the circuit input state.

        :return: The circuit unitary matrix
        :rtype: Tensor
        """
        return self._circ.matrix()

    # def measure_reference(
    #     self, *index: int, with_prob: bool = False
    # ) -> Tuple[str, float]:
    #     return self._circ.measure_reference(*index, with_prob=with_prob)

    def expectation(
        self,
        *ops: Tuple[tn.Node, List[int]],
        reuse: bool = True,
        enable_lightcone: bool = False,
        nmc: int = 1000,
        **kws: Any,
    ) -> Tensor:
        """
        Compute expectation(s) of local operators.

        :param ops: Pairs of `(operator_node, [sites])` specifying where each operator acts.
        :type ops: Tuple[tn.Node, List[int]]
        :param reuse: If True, then the wavefunction tensor is cached for further expectation evaluation,
            defaults to be true.
        :type reuse: bool, optional
        :param enable_lightcone: whether enable light cone simplification, defaults to False
        :type enable_lightcone: bool, optional
        :param nmc: repetition time for Monte Carlo sampling for noisfy calculation, defaults to 1000
        :type nmc: int, optional
        :raises ValueError: "Cannot measure two operators in one index"
        :return: Tensor with one element
        :rtype: Tensor
        """
        return self._circ.expectation(
            *ops,
            reuse=reuse,
            enable_lightcone=enable_lightcone,
            noise_conf=None,
            nmc=nmc,
            **kws,
        )

    def measure_jit(
        self, *index: int, with_prob: bool = False, status: Optional[Tensor] = None
    ) -> Tuple[Tensor, Tensor]:
        """
        Take measurement on the given site indices (computational basis).
        This method is jittable!

        :param index: Measure on which site (wire) index.
        :type index: int
        :param with_prob: If true, theoretical probability is also returned.
        :type with_prob: bool, optional
        :param status: external randomness, with shape [index], defaults to None
        :type status: Optional[Tensor]
        :return: The sample output and probability (optional) of the quantum line.
        :rtype: Tuple[Tensor, Tensor]
        """
        return self._circ.measure_jit(*index, with_prob=with_prob, status=status)

    measure = measure_jit

    def amplitude(self, l: Union[str, Tensor]) -> Tensor:
        r"""
        Return the amplitude for a given base-`d` string `l`.

        For state simulators, this computes :math:`\langle l \vert \psi \rangle`.
        For density-matrix simulators, it would compute :math:`\operatorname{Tr}(\rho \vert l \rangle \langle l \vert)`
        (note the square in magnitude differs between the two formalisms).

        **Example**

        >>> c = tc.QuditCircuit(2, dim=3)
        >>> c.x(0)
        >>> c.x(1)
        >>> c.amplitude("20")
        array(1.+0.j, dtype=complex64)
        >>> c.csum(0, 1, cv=2)
        >>> c.amplitude("21")
        array(1.+0.j, dtype=complex64)

        :param l: Bitstring in base-`d` using `0-9A-Z`.
        :type l: Union[str, Tensor]
        :return: Complex amplitude.
        :rtype: Tensor
        """
        return self._circ.amplitude(l)

    def probability(self) -> Tensor:
        """
        Get the length-`d^n` probability vector over the computational basis.

        :return: Probability vector of shape `[dim^n]`.
        :rtype: Tensor
        """
        return self._circ.probability()

    @partial(arg_alias, alias_dict={"format": ["format_"]})
    def sample(
        self,
        batch: Optional[int] = None,
        allow_state: bool = False,
        readout_error: Optional[Sequence[Any]] = None,
        format: Optional[SAMPLE_FORMAT] = None,
        random_generator: Optional[Any] = None,
        status: Optional[Tensor] = None,
        jittable: bool = True,
    ) -> Any:
        r"""
        Batched sampling from the circuit or final state.

        :param batch: Number of samples. If `None`, returns a single draw.
        :type batch: Optional[int]
        :param allow_state: If `True`, sample from the final state (when memory allows). Prefer `True` for speed.
        :type allow_state: bool
        :param readout_error: Optional readout error model.
        :type readout_error: Optional[Sequence[Any]]
        :param format: Output format. See :py:meth:`tensorcircuit.quantum.measurement_results`.
                       Supported formats for qudits include:

                "count_vector": # np.array([2, 0, 0, 0])

                "count_dict_bin": # {"00": 2, "01": 0, "10": 0, "11": 0}
                    for cases :math:`d\in [11, 36]`, use 0-9A-Z digits (e.g., 'A' -> 10, ..., 'Z' -> 35);

        :type format: Optional[str]
        :param random_generator: random generator,  defaults to None
        :type random_generator: Optional[Any], optional
        :param status: external randomness given by tensor uniformly from [0, 1],
            if set, can overwrite random_generator, shape [batch] for `allow_state=True`
            and shape [batch, nqudits] for `allow_state=False` using perfect sampling implementation
        :type status: Optional[Tensor]
        :param jittable: when converting to count, whether keep the full size. if false, may be conflict
            external jit, if true, may fail for large scale system with actual limited count results
        :type jittable: bool, defaults true
        :return: List (if batch) of tuple (binary configuration tensor and corresponding probability)
            if the format is None, and consistent with format when given
        :rtype: Any
        :raises NotImplementedError: For integer-based output formats not suitable for qudits.
        """
        if format in ["sample_int", "count_tuple", "count_dict_int", "count_vector"]:
            raise NotImplementedError(
                "`int` representation is not friendly for d-dimensional systems."
            )
        return self._circ.sample(
            batch=batch,
            allow_state=allow_state,
            readout_error=readout_error,
            format=format,
            random_generator=random_generator,
            status=status,
            jittable=jittable,
        )

    def projected_subsystem(self, traceout: Tensor, left: Tuple[int, ...]) -> Tensor:
        """
        remaining wavefunction or density matrix on sites in `left`, while fixing the other
        sites to given digits as indicated by `traceout`.

        :param traceout: A tensor encoding digits (0..d-1) for sites to be fixed; jittable.
        :type traceout: Tensor
        :param left: Tuple of site indices to keep (non-jittable argument).
        :type left: Tuple[int, ...]
        :return: Remaining wavefunction or density matrix on the kept sites.
        :rtype: Tensor
        """
        return self._circ.projected_subsystem(
            traceout=traceout,
            left=left,
        )

    def replace_inputs(self, inputs: Tensor) -> None:
        """
        Replace the input state with the circuit structure unchanged.

        :param inputs: Input wavefunction.
        :type inputs: Tensor
        """
        return self._circ.replace_inputs(inputs)

    def mid_measurement(self, index: int, keep: int = 0) -> Tensor:
        """
        Mid-circuit Z-basis post-selection.

        The returned state is **not normalized**; normalize manually if needed.

        :param index: Qudit index where post-selection is applied.
        :type index: int
        :param keep: Post-selected digit in `{0, ..., d-1}`.
        :type keep: int
        :return: Unnormalized post-selected state.
        :rtype: Tensor
        """
        return self._circ.mid_measurement(index, keep=keep)

    mid_measure = mid_measurement
    post_select = mid_measurement
    post_selection = mid_measurement

    def get_quvector(self) -> QuVector:
        """
        Get the representation of the output state in the form of ``QuVector``
        while maintaining the circuit uncomputed

        :return: ``QuVector`` representation of the output state from the circuit
        :rtype: QuVector
        """
        return self._circ.quvector()

    quvector = get_quvector

    def replace_mps_inputs(self, mps_inputs: QuOperator) -> None:
        """
        Replace the input state (keeps circuit structure) using an MPS/MPO-like representation.

        **Example**

        >>> c = tc.QuditCircuit(2, dim=3)
        >>> c.x(0)
        >>> c2 = tc.QuditCircuit(2, dim=3, mps_inputs=c.quvector())
        >>> c2.x(0); c2.wavefunction()
        array([...], dtype=complex64)
        >>> c3 = tc.QuditCircuit(2, dim=3)
        >>> c3.x(0)
        >>> c3.replace_mps_inputs(c.quvector()); c3.wavefunction()
        array([...], dtype=complex64)

        :param mps_inputs: (Nodes, dangling Edges) for a MPS like initial wavefunction.
        :type mps_inputs: Tuple[Sequence[Gate], Sequence[Edge]]
        """
        return self._circ.replace_mps_inputs(mps_inputs)

    def expectation_before(
        self,
        *ops: Tuple[tn.Node, List[int]],
        reuse: bool = True,
        **kws: Any,
    ) -> List[tn.Node]:
        """
        Build (but do not contract) the tensor network for expectation evaluation.

        :param reuse: _description_, defaults to True
        :type reuse: bool, optional
        :raises ValueError: _description_
        :return: _description_
        :rtype: List[tn.Node]
        """
        return self._circ.expectation_before(*ops, reuse=reuse, **kws)

    def amplitude_before(self, l: Union[str, Tensor]) -> List[Gate]:
        r"""
        Return the (uncontracted) tensor network nodes for the amplitude of configuration `l`.

        For state simulators, this corresponds to :math:`\langle l \vert \psi \rangle`.
        For density-matrix simulators,
        it would correspond to :math:`\operatorname{Tr}(\rho \vert l \rangle \langle l \vert)`.

        :param l: Base-`d` string using `0-9A-Z` or an equivalent tensor index.
        :type l: Union[str, Tensor]
        :return: The tensornetwork nodes for the amplitude of the circuit.
        :rtype: List[Gate]
        """
        return self._circ.amplitude_before(l)

    def general_kraus(
        self,
        kraus: Sequence[Gate],
        *index: int,
        status: Optional[float] = None,
        with_prob: bool = False,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Monte Carlo trajectory simulation of general Kraus channel whose Kraus operators cannot be
        amplified to unitary operators. For unitary operators composed Kraus channel, :py:meth:`unitary_kraus`
        is much faster.

        This function is jittable in theory. But only jax+GPU combination is recommended for jit
        since the graph building time is too long for other backend options; though the running
        time of the function is very fast for every case.

        :param kraus: A list of ``tn.Node`` for Kraus operators.
        :type kraus: Sequence[Gate]
        :param index: The qubits index that Kraus channel is applied on.
        :type index: int
        :param status: Random tensor uniformly between 0 or 1, defaults to be None,
            when the random number will be generated automatically
        :type status: Optional[float], optional
        """
        return self._circ.general_kraus(
            kraus,
            *index,
            status=status,
            with_prob=with_prob,
            name=name,
        )

    def unitary_kraus(
        self,
        kraus: Sequence[Gate],
        *index: int,
        prob: Optional[Sequence[float]] = None,
        status: Optional[float] = None,
        name: Optional[str] = None,
    ) -> Tensor:
        """
        Apply unitary gates in ``kraus`` randomly based on corresponding ``prob``.
        If ``prob`` is ``None``, this is reduced to kraus channel language.

        :param kraus: List of ``tc.gates.Gate`` or just Tensors
        :type kraus: Sequence[Gate]
        :param prob: prob list with the same size as ``kraus``, defaults to None
        :type prob: Optional[Sequence[float]], optional
        :param status: random seed between 0 to 1, defaults to None
        :type status: Optional[float], optional
        :return: shape [] int dtype tensor indicates which kraus gate is actually applied
        :rtype: Tensor
        """
        return self._circ.unitary_kraus(
            kraus,
            *index,
            prob=prob,
            status=status,
            name=name,
        )
